// $Id: wish-recv.c,v 1.2 2003/11/19 16:08:22 ensc Exp $    --*- c++ -*--

// Copyright (C) 2003 Enrico Scholz <enrico.scholz@informatik.tu-chemnitz.de>
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; version 2 of the License.
//  
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//  
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.


#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include "wish.h"
#include "wish-internal.h"
#include "parameters.h"
#include "util.h"
#include "util-io.h"
#include "wrappers-io.h"
#include "logging.h"

#include <unistd.h>
#include <sys/socket.h>
#include <assert.h>

inline static ALWAYSINLINE bool
isSamePid(struct Wish const *w, struct ucred const *c)
{
  return w->pid_ == reinterpret_cast(pid_t)(c->pid);
}

static bool
sendWishPleaseAck(struct Wish *self)
{
  Wish_ArgCountType const	argc = 2;
  Wish_ArgSizeType		len  = sizeof WISH_PLEASE_ACK;

  if (!WsendAll(self->fd_, &argc, sizeof argc) ||
      !WsendAll(self->fd_, &len,  sizeof len)  ||
      !WsendAll(self->fd_, WISH_PLEASE_ACK, len))
    RETURN_WITH_LOG("sending PLEASE ACK", false);

  len = sizeof(self->confirm_data_);

  if (!fillRandom(self->confirm_data_, len) ||
      !WsendAll(self->fd_, &len, sizeof len) ||
      !WsendAll(self->fd_, self->confirm_data_, len)) {
    RETURN_WITH_LOG("sending PLEASE ACK confirmation", false);
  }

  return true;
}

static bool
recvWishAck(struct Wish *self)
{
  Wish_ArgSizeType	len;
  struct ucred		creds;

  assert(self->has_creds_);

  if (!readCreds(self->fd_, &len, sizeof len, &creds) ||
      len!=sizeof(self->confirm_data_) ||
      !isSamePid(self, &creds) ||
      self->creds_.ctx!=getContextOfPid(creds.pid))
    RETURN_WITH_LOG("receiving ACK", false);

  char			buf[len];

  if (!readCreds(self->fd_, buf, len, &creds) ||
      !isSamePid(self, &creds) ||
      memcmp(buf, self->confirm_data_, sizeof buf)!=0)
    RETURN_WITH_LOG("verifying ACK", false);

  return true;
}

static bool
readVerified(struct Wish *self,
	     struct Parameters *params)
{
  Wish_ArgSizeType	len;
  struct ucred		creds;

  if (!readCreds(self->fd_, &len, sizeof len, &creds) ||
      !isSamePid(self, &creds)) {
    RETURN_WITH_LOG("checking sender", false);
  }

  char		*tmp = Parameters_allocParam(params, len);
  if (len>0 &&
      ((tmp==0 ||
	!readCreds(self->fd_, tmp, len, &creds) ||
	!isSamePid(self, &creds)))) {
    RETURN_WITH_LOG("allocating data", false);
  }

  return true;
}

static bool
readVerifiedN(struct Wish *self,
	      struct Parameters *params, size_t cnt)
{
  for (size_t i=0; i<cnt; ++i)
    if (!readVerified(self, params)) return false;

  return Parameters_normalize(params);
}

bool
Wish_recv(struct Wish *self, struct Parameters *params)
{
  Wish_ArgCountType	cnt;
  struct ucred		creds;

  if (!readCreds(self->fd_, &cnt, sizeof(cnt), &creds))
    RETURN_WITH_LOG("reading credentials", false);

  ctx_t			ctx = getContextOfPid(creds.pid);
  if (!self->has_creds_) {
    self->creds_.ctx = ctx;
    self->creds_.uid = creds.uid;
    self->creds_.gid = creds.gid;
    self->has_creds_ = true;
  }

  self->pid_ = creds.pid;

  if (ctx!=self->creds_.ctx ||
      !Parameters_setArgCount(params, cnt) ||
      !readVerifiedN(self, params, cnt) ||
      !sendWishPleaseAck(self) ||
      !recvWishAck(self)) {
    RETURN_WITH_LOG("reading command", false);
  }

  return true;
}

struct Credentials const *
Wish_getCredentials(struct Wish const *self)
{
  if (self->has_creds_) return &self->creds_;
  else                  return 0;
}

bool
Wish_sendResult(int fd, unsigned int code_i, char const *data, size_t len)
{
  uint16_t		code    = code_i;
  Wish_ArgSizeType	tot_len = sizeof(code);

  if (len==0) len=strlen(data);
  tot_len += len;

  if (!WsendAll(fd, &tot_len, sizeof tot_len) ||
      !WsendAll(fd, &code,    sizeof code) ||
      !WsendAll(fd, data,     len))
    RETURN_WITH_LOG("sending return code", false);

  return true;
}
